@@ -18,6 +18,7 @@ related utilities used by hledger commands.
 module Hledger.Cli.CliOptions (
   progname,
   prognameandversion,
+  binaryinfo,
 
   -- * cmdargs flags & modes
   inputflags,
@@ -49,7 +50,6 @@ module Hledger.Cli.CliOptions (
   showModeUsage,
   withAliases,
   likelyExecutablesInPath,
-  hledgerExecutablesInPath,
 
   -- * CLI options
   CliOpts(..),
@@ -58,6 +58,8 @@ module Hledger.Cli.CliOptions (
   getHledgerCliOpts,
   getHledgerCliOpts',
   rawOptsToCliOpts,
+  cliOptsDropArgs,
+  argsAddDoubleDash,
   outputFormats,
   defaultOutputFormat,
   CommandHelpStr,
@@ -67,19 +69,20 @@ module Hledger.Cli.CliOptions (
   -- * CLI option accessors
   -- | These do the extra processing required for some options.
   journalFilePathFromOpts,
+  journalFilePathFromOptsNoDefault,
   rulesFilePathFromOpts,
   outputFileFromOpts,
   outputFormatFromOpts,
   defaultWidth,
-  -- widthFromOpts,
   replaceNumericFlags,
   ensureDebugFlagHasVal,
   -- | For register:
   registerWidthsFromOpts,
 
   -- * Other utils
-  hledgerAddons,
   topicForMode,
+  UsedOrDeclared(..),
+  usedOrDeclaredFromOpts,
 
 --  -- * Convenience re-exports
 --  module Data.String.Here,
@@ -87,18 +90,17 @@ module Hledger.Cli.CliOptions (
 )
 where
 
-import qualified Control.Exception as C
+import Control.Exception qualified as C
 import Control.Monad (when)
 import Data.Char
 import Data.Default
-import Data.Either (isRight)
-import Data.List.Extra (groupSortOn, intercalate, isInfixOf, nubSort)
-import qualified Data.List.NonEmpty as NE (NonEmpty, fromList, head, nonEmpty)
+import Data.List.Extra (intercalate, isInfixOf, nubSort)
+import Data.List.NonEmpty qualified as NE (NonEmpty, fromList, nonEmpty)
 import Data.List.Split (splitOn)
 import Data.Maybe
 --import Data.String.Here
 -- import Data.Text (Text)
-import qualified Data.Text as T
+import Data.Text qualified as T
 import Data.Void (Void)
 import GitHash (tGitInfoCwdTry)
 import Safe
@@ -106,9 +108,6 @@ import String.ANSI
 import System.Console.CmdArgs hiding (Default,def)
 import System.Console.CmdArgs.Explicit
 import System.Console.CmdArgs.Text
-#ifndef mingw32_HOST_OS
-import System.Console.Terminfo
-#endif
 import System.Directory
 import System.Environment
 import System.Exit (exitSuccess)
@@ -142,13 +141,27 @@ prognameandversion =
   progname
   packageversion
 
+binaryinfo :: HledgerBinaryInfo
+Right binaryinfo = parseHledgerVersion prognameandversion
+-- ui and web use nullbinaryinfo for a parse failure here to silence an inexhaustive pattern warning.
+-- I can't reproduce that warning right now, so here I've stuck with the original approach,
+-- which will force a compile error if prognameandversion is ever malformed, eg from unexpected
+-- git output.
+
 -- Common options.
 -- keep synced: the docs macro in doc/common.m4
 
 -- | Common input-related flags: --file, --rules, --conf, --alias...
 inputflags :: [Flag RawOpts]
 inputflags = [
-   flagReq  ["file","f"]      (\s opts -> Right $ setopt "file" s opts) "[FMT:]FILE" "Read data from FILE, or from stdin if FILE is -, inferring format from extension or a FMT: prefix. Can be specified more than once. If not specified, reads from $LEDGER_FILE or $HOME/.hledger.journal."
+   flagReq  ["file","f"]      (\s opts -> Right $ setopt "file" s opts) "[FMT:]FILE" $ unwords [
+     "Use this as the journal file (- means stdin)."
+    ,"If not specified, $LEDGER_FILE or ~/.hledger.journal will be used."
+    ,"If specified more than once, the files will be read in order."
+    ,"Each file's format (journal, csv, timeclock, timedot, rules..) is inferred from the file extension or a FMT: prefix."
+    ,"Some commands (add, import) write to the (first) file,"
+    ,"and expect it to be in journal format."
+    ]
   ,flagReq  ["rules"]         (\s opts -> Right $ setopt "rules" s opts) "RULESFILE" "Use rules defined in this rules file for converting subsequent CSV/SSV/TSV files. If not specified, uses FILE.csv.rules for each FILE.csv."  -- see also hiddenflags
 
   ,flagReq  ["alias"]         (\s opts -> Right $ setopt "alias" s opts)  "A=B|/RGX/=RPL" "transform account names from A to B, or by replacing regular expression matches"
@@ -160,6 +173,11 @@ inputflags = [
     , "In hledger-ui, also make future-dated transactions visible at startup."
     ])
   ,flagNone ["ignore-assertions","I"] (setboolopt "ignore-assertions") "don't check balance assertions by default"
+  ,flagReq  ["txn-balancing"] (\s opts -> Right $ setopt "txn-balancing" s opts) "..." (unlines [
+     "how to check that transactions are balanced:"
+    ,"'old':   - use global display precision"
+    ,"'exact': - use transaction precision (default)"
+    ])
   ,flagNone ["infer-costs"] (setboolopt "infer-costs") "infer conversion equity postings from costs"
   ,flagNone ["infer-equity"] (setboolopt "infer-equity") "infer costs from conversion equity postings"
   -- history of this flag so far, lest we be confused:
@@ -187,11 +205,11 @@ reportflags = [
   -- report period, interval, dates
   flagReq  ["begin","b"]     (\s opts -> Right $ setopt "begin" s opts) "DATE" "include postings/transactions on/after this date"
  ,flagReq  ["end","e"]       (\s opts -> Right $ setopt "end" s opts) "DATE" "include postings/transactions before this date (with a report interval, will be adjusted to following subperiod end)"
- ,flagNone ["daily","D"]     (setboolopt "daily")     "multiperiod report with 1 day interval"
- ,flagNone ["weekly","W"]    (setboolopt "weekly")    "multiperiod report with 1 week interval"
- ,flagNone ["monthly","M"]   (setboolopt "monthly")   "multiperiod report with 1 month interval"
- ,flagNone ["quarterly","Q"] (setboolopt "quarterly") "multiperiod report with 1 quarter interval"
- ,flagNone ["yearly","Y"]    (setboolopt "yearly")    "multiperiod report with 1 year interval"
+ ,flagNone ["daily","D"]     (setboolopt "daily")     "set report interval: 1 day"
+ ,flagNone ["weekly","W"]    (setboolopt "weekly")    "set report interval: 1 week"
+ ,flagNone ["monthly","M"]   (setboolopt "monthly")   "set report interval: 1 month"
+ ,flagNone ["quarterly","Q"] (setboolopt "quarterly") "set report interval: 1 quarter"
+ ,flagNone ["yearly","Y"]    (setboolopt "yearly")    "set report interval: 1 year"
  ,flagReq  ["period","p"]    (\s opts -> Right $ setopt "period" s opts) "PERIODEXP" "set begin date, end date, and/or report interval, with more flexibility"
  ,flagReq  ["today"]         (\s opts -> Right $ setopt "today" s opts) "DATE" "override today's date (affects relative dates)"
  ,flagNone ["date2"]         (setboolopt "date2") "match/use secondary dates instead (deprecated)"  -- see also hiddenflags
@@ -205,20 +223,21 @@ reportflags = [
  ,flagReq  ["depth"]         (\s opts -> Right $ setopt "depth" s opts) "DEPTHEXP" "if a number (or -NUM): show only top NUM levels of accounts. If REGEXP=NUM, only apply limiting to accounts matching the regular expression."
 
   -- valuation
- ,flagNone ["B","cost"]      (setboolopt "B") "show amounts converted to their cost/sale amount"
+ ,flagNone ["B","cost"]      (setboolopt "B") "convert amounts to their cost/sale amount (@/@@)"
+    -- ^ no "valuation mode:" prefix for this one, it's not mutually exclusive
  ,flagNone ["V","market"]    (setboolopt "V")
-    (unlines
-      ["Show amounts converted to their value at period end(s) in their default valuation commodity."
-      ,"Equivalent to --value=end."
+    (unwords
+      [valuationprefix ++ "show amounts converted to market value at period end(s) in their default valuation commodity."
+      ,"Short for --value=end."
       ])
  ,flagReq ["X","exchange"]   (\s opts -> Right $ setopt "X" s opts) "COMM"
-    (unlines
-      ["Show amounts converted to their value at period end(s) in the specified commodity."
-      ,"Equivalent to --value=end,COMM."
+    (unwords
+      [valuationprefix ++ "show amounts converted to market value at period end(s) in the specified commodity."
+      ,"Short for --value=end,COMM."
       ])
  ,flagReq  ["value"]         (\s opts -> Right $ setopt "value" s opts) "WHEN[,COMM]"
     (unlines
-      ["show amounts converted to their value on the specified date(s) in their default valuation commodity or a specified commodity. WHEN can be:"
+      [valuationprefix ++ "show amounts converted to market value on the specified date(s) in their default valuation commodity or a specified commodity. WHEN can be:"
       ,"'then':     value on transaction dates"
       ,"'end':      value at period end(s)"
       ,"'now':      value today"
@@ -229,8 +248,10 @@ reportflags = [
  ,flagReq ["commodity-style", "c"] (\s opts -> Right $ setopt "commodity-style" s opts) "S"
     "Override a commodity's display style.\nEg: -c '$1000.' or -c '1.000,00 EUR'"
  ,flagOpt "yes" ["pretty"] (\s opts -> Right $ setopt "pretty" s opts) "YN"
-    "Use box-drawing characters in text output? Can be\n'y'/'yes' or 'n'/'no'.\nIf YN is specified, the equals is required."
+    "Use box-drawing characters in text output? The optional 'y'/'yes' or 'n'/'no' arg requires =."
  ]
+  where
+    valuationprefix = "valuation mode: "
 
 helpflags :: [Flag RawOpts]
 helpflags = [
@@ -261,11 +282,13 @@ terminalflags = [
 flattreeflags :: Bool -> [Flag RawOpts]
 flattreeflags showamounthelp = [
    flagNone ["flat","l"] (setboolopt "flat")
-     ("show accounts as a flat list (default)"
+     (prefix ++ "show accounts as a flat list (default)"
       ++ if showamounthelp then ". Amounts exclude subaccount amounts, except where the account is depth-clipped." else "")
   ,flagNone ["tree","t"] (setboolopt "tree")
-    ("show accounts as a tree" ++ if showamounthelp then ". Amounts include subaccount amounts." else "")
+    (prefix ++ "show accounts as a tree" ++ if showamounthelp then ". Amounts include subaccount amounts." else "")
   ]
+  where
+    prefix = "list/tree mode: "
 
 -- | hledger CLI's --conf/--no-conf flags.
 confflags = [
@@ -282,6 +305,8 @@ hiddenflagsformainmode = [
   ,flagNone ["pretty-tables"]        (setopt "pretty" "always") "legacy flag that was renamed"
   ,flagNone ["anon"]                 (setboolopt "anon") "deprecated, renamed to --obfuscate"  -- #2133, handled by anonymiseByOpts
   ,flagNone ["obfuscate"]            (setboolopt "obfuscate") "slightly obfuscate hledger's output. Warning, does not give privacy. Formerly --anon."  -- #2133, handled by maybeObfuscate
+  ,flagNone ["old-timeclock", "timeclock-old"] (setboolopt "oldtimeclock") "don't pair timeclock entries by account name"
+  ,flagNone ["old-glob"]             (setboolopt "oldglob") "deprecated, no longer used as of 1.50.4"  -- #2498
   ,flagReq  ["rules-file"]           (\s opts -> Right $ setopt "rules" s opts) "RULESFILE" "was renamed to --rules"
   ]
 
@@ -332,8 +357,8 @@ mkgeneralflagsgroups3 helpflags' = [
    (helpflagstitle, helpflags')
   ]
 
-inputflagstitle  = "\nGeneral input/data transformation flags"
-outputflagstitle = "\nGeneral output/reporting flags (supported by some commands)"
+inputflagstitle  = "\nGeneral input flags"
+outputflagstitle = "\nGeneral output flags (affecting some commands)"
 helpflagstitle   = "\nGeneral help flags"
 
 -- Several subsets of the standard general flags plus CLI help flags, as separate groups.
@@ -433,7 +458,7 @@ hledgerCommandMode :: CommandHelpStr -> [Flag RawOpts] -> [(String, [Flag RawOpt
   -> [Flag RawOpts] -> ([Arg RawOpts], Maybe (Arg RawOpts)) -> Mode RawOpts
 hledgerCommandMode helpstr unnamedflaggroup namedflaggroups hiddenflaggroup argsdescr =
   case parseCommandHelp helpstr of
-    Nothing -> error' $ "Could not parse command doc:\n"++helpstr++"\n"  -- PARTIAL:
+    Nothing -> error' $ "could not parse command doc:\n"++helpstr++"\n"  -- PARTIAL:
     Just CommandHelp{cmdName, mcmdShortName, cmdHelpPreamble, cmdHelpPostamble} ->
       (defCommandMode $ cmdName : maybeToList mcmdShortName) {
          modeHelp        = cmdHelpPreamble
@@ -551,9 +576,8 @@ data CliOpts = CliOpts {
     ,no_new_accounts_ :: Bool           -- add
     ,width_           :: Maybe String   -- ^ the --width value provided, if any
     ,available_width_ :: Int            -- ^ estimated usable screen width, based on
-                                        -- 1. the COLUMNS env var, if set
-                                        -- 2. the width reported by the terminal, if supported
-                                        -- 3. the default (80)
+                                        -- 1. the width reported by the terminal, if supported
+                                        -- 2. the default (80)
     ,progstarttime_   :: POSIXTime      -- system POSIX time at start
  } deriving (Show)
 
@@ -616,15 +640,8 @@ rawOptsToCliOpts rawopts = do
   usecolor <- useColorOnStdout
   let iopts = rawOptsToInputOpts day usecolor postingaccttags rawopts
   rspec <- either error' pure $ rawOptsToReportSpec day usecolor rawopts  -- PARTIAL:
-  mcolumns <- readMay <$> getEnvSafe "COLUMNS"
-  mtermwidth <-
-#ifdef mingw32_HOST_OS
-    return Nothing
-#else
-    (`getCapability` termColumns) <$> setupTermFromEnv
-    -- XXX Throws a SetupTermError if the terminfo database could not be read, should catch
-#endif
-  let availablewidth = NE.head $ NE.fromList $ catMaybes [mcolumns, mtermwidth, Just defaultWidth]  -- PARTIAL: fromList won't fail because non-null list
+  mtermwidth <- getTerminalWidth
+  let availablewidth = fromMaybe defaultWidth mtermwidth
   return defcliopts {
               rawopts_         = rawopts
              ,command_         = command
@@ -641,6 +658,18 @@ rawOptsToCliOpts rawopts = do
              ,available_width_ = availablewidth
              }
 
+-- | Drop the arguments ("args") from this CliOpts' rawopts field.
+cliOptsDropArgs :: CliOpts -> CliOpts
+cliOptsDropArgs copts@CliOpts{rawopts_} = copts{rawopts_ = dropRawOpt "args" rawopts_}
+
+-- | cmdargs eats the first double-dash (--) argument when parsing a command line,
+-- which causes problems for the run and repl commands.
+-- Sometimes we work around this by duplicating that first -- argument.
+-- This doesn't break anything that we know of yet.
+argsAddDoubleDash args'
+  | "--" `elem` args' = let (as,bs) = break (=="--") args' in as <> ["--"] <> bs
+  | otherwise = args'
+
 -- | A helper for addon commands: this parses options and arguments from
 -- the current command line using the given hledger-style cmdargs mode,
 -- and returns a CliOpts. Or, with --help or -h present, it prints
@@ -706,12 +735,20 @@ getHledgerCliOpts mode' = do
 -- File paths can have a READER: prefix naming a reader/data format.
 journalFilePathFromOpts :: CliOpts -> IO (NE.NonEmpty String)
 journalFilePathFromOpts opts = do
-  f <- defaultJournalPath
+  mbpaths <- journalFilePathFromOptsNoDefault opts
+  case mbpaths of
+    Just paths -> return paths
+    Nothing -> do
+      f <- defaultJournalPath
+      return $ NE.fromList [f]
+
+-- | Like journalFilePathFromOpts, but does not use defaultJournalPath
+journalFilePathFromOptsNoDefault :: CliOpts -> IO (Maybe (NE.NonEmpty String))
+journalFilePathFromOptsNoDefault opts = do
   d <- getCurrentDirectory
-  maybe
-    (return $ NE.fromList [f])
-    (mapM (expandPathPreservingPrefix d))
-    $ NE.nonEmpty $ file_ opts
+  case NE.nonEmpty $ file_ opts of
+    Nothing -> return Nothing
+    Just paths -> Just <$> mapM (expandPathPreservingPrefix d) paths
 
 expandPathPreservingPrefix :: FilePath -> PrefixedFilePath -> IO PrefixedFilePath
 expandPathPreservingPrefix d prefixedf = do
@@ -765,24 +802,11 @@ rulesFilePathFromOpts opts = do
   d <- getCurrentDirectory
   maybe (return Nothing) (fmap Just . expandPath d) $ mrules_file_ $ inputopts_ opts
 
--- -- | Get the width in characters to use for console output.
--- -- This comes from the --width option, or the COLUMNS environment
--- -- variable, or (on posix platforms) the current terminal width, or 80.
--- -- Will raise a parse error for a malformed --width argument.
--- widthFromOpts :: CliOpts -> Int
--- widthFromOpts CliOpts{width_=Nothing, available_width_=w} = w
--- widthFromOpts CliOpts{width_=Just s}  =
---     case runParser (read `fmap` some digitChar <* eof :: ParsecT Void String Identity Int) "(unknown)" s of
---         Left e   -> usageError $ "could not parse width option: "++errorBundlePretty e
---         Right w  -> w
-
--- for register:
-
 -- | Get the width in characters to use for the register command's console output,
 -- and also the description column width if specified (following the main width, comma-separated).
 -- The widths will be as follows:
 -- @
--- no --width flag - overall width is the available width (COLUMNS, or posix terminal width, or 80); description width is unspecified (auto)
+-- no --width flag - overall width is the available width (or terminal width, or 80); description width is unspecified (auto)
 -- --width W       - overall width is W, description width is auto
 -- --width W,D     - overall width is W, description width is D
 -- @
@@ -801,50 +825,31 @@ registerWidthsFromOpts CliOpts{width_=Just s}  =
           eof
           return (totalwidth, descwidth)
 
+-- A common choice for filtering lists of declarable things.
+data UsedOrDeclared
+  = Used
+  | Declared
+  | Undeclared
+  | Unused
+  deriving (Show, Eq)
+
+-- Get the flag of this kind from opts, or raise an error if there's more than one.
+usedOrDeclaredFromOpts :: CliOpts -> Maybe UsedOrDeclared
+usedOrDeclaredFromOpts CliOpts{rawopts_=rawopts} =
+  case ( boolopt "used"       rawopts
+       , boolopt "declared"   rawopts
+       , boolopt "undeclared" rawopts
+       , boolopt "unused"     rawopts
+       ) of
+    (False, False, False, False) -> Nothing
+    (True, False, False, False) -> Just Used
+    (False, True, False, False) -> Just Declared
+    (False, False, True, False) -> Just Undeclared
+    (False, False, False, True) -> Just Unused
+    _ -> error' "please pick at most one of --used, --declared, --undeclared, --unused"
+
 -- Other utils
 
--- | Get the sorted unique canonical names of hledger addon commands
--- found in the current user's PATH. These are used in command line
--- parsing and to display the commands list.
---
--- Canonical addon names are the filenames of hledger-* executables in
--- PATH, without the "hledger-" prefix, and without the file extension
--- except when it's needed for disambiguation (see below).
---
--- When there are exactly two versions of an executable (same base
--- name, different extensions) that look like a source and compiled
--- pair (one has .exe, .com, or no extension), the source version will
--- be excluded (even if it happens to be newer). When there are three
--- or more versions (or two versions that don't look like a
--- source/compiled pair), they are all included, with file extensions
--- intact.
---
-hledgerAddons :: IO [String]
-hledgerAddons = do
-  -- past bug generator
-  as1 <- hledgerExecutablesInPath                     -- ["hledger-check","hledger-check-dates","hledger-check-dates.hs","hledger-check.hs","hledger-check.py"]
-  let as2 = map stripPrognamePrefix as1               -- ["check","check-dates","check-dates.hs","check.hs","check.py"]
-  let as3 = groupSortOn takeBaseName as2              -- [["check","check.hs","check.py"],["check-dates","check-dates.hs"]]
-  let as4 = concatMap dropRedundantSourceVersion as3  -- ["check","check.hs","check.py","check-dates"]
-  return as4
-
-stripPrognamePrefix = drop (length progname + 1)
-
-dropRedundantSourceVersion [f,g]
-  | map toLower (takeExtension f) `elem` compiledExts = [f]
-  | map toLower (takeExtension g) `elem` compiledExts = [g]
-dropRedundantSourceVersion fs = fs
-
-compiledExts = ["",".com",".exe"]
-
--- | Get the sorted unique filenames of all hledger-* executables in
--- the current user's PATH. These are files in any of the PATH directories,
--- named hledger-*, with either no extension (and no periods in the name)
--- or one of the addonExtensions.
--- We do not currently filter out non-file objects or files without execute permission.
-hledgerExecutablesInPath :: IO [String]
-hledgerExecutablesInPath = filter isHledgerExeName <$> likelyExecutablesInPath
-
 -- None of https://hackage.haskell.org/package/directory-1.3.8.1/docs/System-Directory.html#g:5
 -- do quite what we need (find all the executables in PATH with a filename prefix).
 -- | Get all sorted unique filenames in the current user's PATH.
@@ -865,37 +870,6 @@ likelyExecutablesInPath = do
 -- return exes''
 -- where isExecutable f = getPermissions f >>= (return . executable)
 
-isHledgerExeName :: String -> Bool
-isHledgerExeName = isRight . parsewith hledgerexenamep . T.pack
-    where
-      hledgerexenamep = do
-        _ <- string $ T.pack progname
-        _ <- char '-'
-        _ <- some $ noneOf ['.']
-        optional (string "." >> choice' (map (string . T.pack) addonExtensions))
-        eof
-
--- stripAddonExtension :: String -> String
--- stripAddonExtension = regexReplace re "" where re = "\\.(" ++ intercalate "|" addonExtensions ++ ")$"
-
-addonExtensions :: [String]
-addonExtensions =
-  ["bat"
-  ,"com"
-  ,"exe"
-  ,"hs"
-  ,"js"
-  ,"lhs"
-  ,"lua"
-  ,"php"
-  ,"pl"
-  ,"py"
-  ,"rb"
-  ,"rkt"
-  ,"sh"
-  -- ,""
-  ]
-
 getEnvSafe :: String -> IO String
 getEnvSafe v = getEnv v `C.catch` (\(_::C.IOException) -> return "") -- XXX should catch only isDoesNotExistError e
 
@@ -903,18 +877,6 @@ getDirectoryContentsSafe :: FilePath -> IO [String]
 getDirectoryContentsSafe d =
     (filter (not . (`elem` [".",".."])) `fmap` getDirectoryContents d) `C.catch` (\(_::C.IOException) -> return [])
 
--- not used:
--- -- | Print debug info about arguments and options if --debug is present.
--- debugArgs :: [String] -> CliOpts -> IO ()
--- debugArgs args opts =
---   when ("--debug" `elem` args) $ do
---     progname <- getProgName
---     putStrLn $ "running: " ++ progname
---     putStrLn $ "raw args: " ++ show args
---     putStrLn $ "processed opts:\n" ++ show opts
---     d <- getCurrentDay
---     putStrLn $ "search query: " ++ (show $ queryFromOpts d $ reportopts_ opts)
-
 -- ** Lenses
 
 makeHledgerClassyLenses ''CliOpts
